/**
  *  \file t_ptrmap.cc
  *
  *  (c) 2001-2015 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "cpluslib/ptrmap.h"

#include "tests.h"

namespace {
    class Obj {
        Obj(const Obj&);
        Obj& operator=(const Obj&);
     public:
        int n;
        static int live;
        Obj(int n)
            : n(n) { ++live; }
        ~Obj()
            { --live; }
    };
    int Obj::live = 0;
}

void
TestPtrMap::test1()
{
    const Obj* null = 0;
    // Blank map
    ptr_map<int,Obj> pm;
    TS_ASSERT(pm.empty());
    TS_ASSERT_EQUALS(pm.size(), 0);
    TS_ASSERT_EQUALS(Obj::live, 0);
    TS_ASSERT_EQUALS(pm.begin(), pm.end());

    // Add three objects
    pm.insert(42, new Obj(1));
    pm.insert(23, new Obj(2));
    pm.insert(95, new Obj(3));
    TS_ASSERT(!pm.empty());
    TS_ASSERT_EQUALS(pm.size(), 3);
    TS_ASSERT_EQUALS(Obj::live, 3);
    TS_ASSERT_EQUALS(pm[42]->n, 1);
    TS_ASSERT_EQUALS(pm[23]->n, 2);
    TS_ASSERT_EQUALS(pm[95]->n, 3);
    TS_ASSERT_EQUALS(pm[17], null);

    // Iteration
    {
        ptr_map<int,Obj>::iterator i = pm.begin();
        TS_ASSERT_DIFFERS(i, pm.end());
        TS_ASSERT_EQUALS(i->first, 23);
        TS_ASSERT_EQUALS(i->second->n, 2);
        ++i;
        TS_ASSERT_DIFFERS(i, pm.end());
        TS_ASSERT_EQUALS(i->first, 42);
        TS_ASSERT_EQUALS(i->second->n, 1);
        ++i;
        TS_ASSERT_DIFFERS(i, pm.end());
        TS_ASSERT_EQUALS(i->first, 95);
        TS_ASSERT_EQUALS(i->second->n, 3);
        ++i;
        TS_ASSERT_EQUALS(i, pm.end());
        --i;
        TS_ASSERT_DIFFERS(i, pm.end());
        TS_ASSERT_EQUALS(i->first, 95);
        TS_ASSERT_EQUALS(i->second->n, 3);
    }

    // Remove one
    int n = pm.erase(12);
    TS_ASSERT_EQUALS(n, 0);
    n = pm.erase(95);
    TS_ASSERT_EQUALS(n, 1);
    TS_ASSERT_EQUALS(pm.size(), 2);
    TS_ASSERT_EQUALS(Obj::live, 2);
    TS_ASSERT_EQUALS(pm[42]->n, 1);
    TS_ASSERT_EQUALS(pm[23]->n, 2);
    TS_ASSERT_EQUALS(pm[95], null);

    // Remove all
    pm.erase(pm.begin(), pm.end());
    TS_ASSERT(pm.empty());
    TS_ASSERT_EQUALS(pm.size(), 0);
    TS_ASSERT_EQUALS(Obj::live, 0);
    TS_ASSERT_EQUALS(pm.begin(), pm.end());

    // Add some more
    pm.insert(7,  new Obj(4));
    pm.insert(8,  new Obj(5));
    pm.insert(9,  new Obj(6));
    pm.insert(10, new Obj(7));
    TS_ASSERT_EQUALS(Obj::live, 4);
    TS_ASSERT_EQUALS(pm[7]->n, 4);
    TS_ASSERT_EQUALS(pm[8]->n, 5);
    TS_ASSERT_EQUALS(pm[9]->n, 6);
    TS_ASSERT_EQUALS(pm[10]->n, 7);

    // Replace
    pm.insert(8, new Obj(8));
    TS_ASSERT_EQUALS(Obj::live, 4);
    TS_ASSERT_EQUALS(pm[7]->n, 4);
    TS_ASSERT_EQUALS(pm[8]->n, 8);
    TS_ASSERT_EQUALS(pm[9]->n, 6);
    TS_ASSERT_EQUALS(pm[10]->n, 7);

    // Swap some
    pm.swap_elems(pm.find(7), pm.find(9));
    TS_ASSERT_EQUALS(Obj::live, 4);
    TS_ASSERT_EQUALS(pm[7]->n, 6);
    TS_ASSERT_EQUALS(pm[8]->n, 8);
    TS_ASSERT_EQUALS(pm[9]->n, 4);
    TS_ASSERT_EQUALS(pm[10]->n, 7);

    // Extract
    Obj* o = pm.extract_elem(pm.find(10));
    TS_ASSERT_DIFFERS(o, null);
    TS_ASSERT_EQUALS(o->n, 7);
    TS_ASSERT_EQUALS(Obj::live, 4);
    delete o;
    TS_ASSERT_EQUALS(Obj::live, 3);
    TS_ASSERT_EQUALS(pm.find(10), pm.end());

    // Temporary map
    {
        ptr_map<int,Obj> tm;
        tm.insert(0, new Obj(22));
        tm.insert(99, 0);
        tm.insert(42, new Obj(17));
        TS_ASSERT_EQUALS(Obj::live, 5);
    }
    TS_ASSERT_EQUALS(Obj::live, 3);

    // Swap
    ptr_map<int,Obj> pm2;
    pm2.insert(8, new Obj(15));
    TS_ASSERT_EQUALS(pm2.size(), 1);
    pm2.swap(pm);
    TS_ASSERT_EQUALS(pm2.size(), 3);
    TS_ASSERT_EQUALS(pm2[7]->n, 6);
    TS_ASSERT_EQUALS(pm2[8]->n, 8);
    TS_ASSERT_EQUALS(pm2[9]->n, 4);

    TS_ASSERT_EQUALS(pm.size(), 1);
    TS_ASSERT_EQUALS(pm[8]->n, 15);

    // Clear
    pm.clear();
    TS_ASSERT_EQUALS(pm.size(), 0);
    TS_ASSERT(pm.empty());
    TS_ASSERT_EQUALS(Obj::live, 3);

    // Count
    TS_ASSERT_EQUALS(pm.count(0), 0);
    TS_ASSERT_EQUALS(pm.count(7), 0);
    TS_ASSERT_EQUALS(pm.count(10), 0);
    TS_ASSERT_EQUALS(pm2.count(0), 0);
    TS_ASSERT_EQUALS(pm2.count(7), 1);
    TS_ASSERT_EQUALS(pm2.count(8), 1);

    // Erase iterator
    pm2.erase(pm2.begin());
    TS_ASSERT_EQUALS(pm2.size(), 2);
    TS_ASSERT_EQUALS(Obj::live, 2);
    TS_ASSERT_EQUALS(pm2[7], null);
    TS_ASSERT_EQUALS(pm2[8]->n, 8);
}
